// // The WASM library
// #[cfg(feature = "wasm")]
// use wasm_bindgen::prelude::*;
// #[cfg(feature = "wasm")]
// use serde::{Deserialize, Serialize};

// #[cfg(feature = "wasm")]
// use crate::{
//     Move, State,
// };

// #[cfg_attr(feature = "wasm", wasm_bindgen)]
// #[cfg(feature = "wasm")]
// pub struct Engine {}

// #[cfg_attr(feature = "wasm", wasm_bindgen)]
// #[cfg(feature = "wasm")]
// impl Engine {
//     pub fn new() -> Self {
//         Engine {}
//     }
